@@ -1,5 +1,5 @@
 /* This file is part of GNU Radius.
-   Copyright (C) 2000,2001,2002,2003,2004 Free Software Foundation, Inc.
+   Copyright (C) 2000,2001,2002,2003,2004,2005 Free Software Foundation, Inc.
 
    Written by Sergey Poznyakoff
  
@@ -13,9 +13,10 @@
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
-   You should have received a copy of the GNU General Public License
-   along with GNU Radius; if not, write to the Free Software Foundation, 
-   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
+   You should have received a copy of the GNU General Public
+   License along with GNU Radius; if not, write to the Free
+   Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
+   Boston, MA 02110-1301 USA. */
 
 #ifdef HAVE_CONFIG_H
 # include <config.h>
@@ -58,7 +59,7 @@ grad_client_message_id(grad_server_t *server)
 	int fd;
 	unsigned id;
 	
-	fd = open(radmsgid_path, O_RDWR|O_CREAT, PERM);
+	fd = open(grad_msgid_file, O_RDWR|O_CREAT, PERM);
 	if (fd != -1) {
 		struct stat st;
 		
@@ -332,8 +333,8 @@ grad_client_send0(grad_server_queue_t *config, int port_type, int code,
 
                                 if (result > 0) 
                                         req = grad_client_recv(
-						sin->sin_addr.s_addr,
-                                                sin->sin_port,
+						ntohl(sin->sin_addr.s_addr),
+                                                ntohs(sin->sin_port),
                                                 server->secret,
                                                 authenticator,
                                                 recv_buf,
@@ -472,7 +473,7 @@ grad_client_create_queue(int read_cfg, grad_uint32_t source_ip, size_t bufsize)
         client->servers = 0;
 
         if (read_cfg) {
-                filename = grad_mkfilename(radius_dir, "client.conf");
+                filename = grad_mkfilename(grad_config_dir, "client.conf");
                 grad_read_raddb_file(filename, 1, NULL,
                                      parse_client_config, client);
                 grad_free(filename);
