@@ -1,5 +1,5 @@
 /* This file is part of GNU Radius.
-   Copyright (C) 2000,2001,2002,2003,2004 Free Software Foundation, Inc.
+   Copyright (C) 2000,2001,2002,2003,2004,2005 Free Software Foundation, Inc.
 
    Written by Sergey Poznyakoff
  
@@ -13,9 +13,10 @@
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
   
-   You should have received a copy of the GNU General Public License
-   along with GNU Radius; if not, write to the Free Software Foundation,
-   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
+   You should have received a copy of the GNU General Public
+   License along with GNU Radius; if not, write to the Free
+   Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
+   Boston, MA 02110-1301 USA. */
 
 #ifdef HAVE_CONFIG_H
 # include <config.h>
@@ -56,23 +57,30 @@ static grad_list_t /* of grad_dict_value_t */ *dictionary_values;
 static grad_list_t /* of grad_dict_vendor_t */ *dictionary_vendors;
 static int         vendorno;
 
+
+static grad_dict_attr_t *
+dict_symbol_ptr(DICT_SYMBOL *sym)
+{
+	switch (sym->type) {
+	case dict_symbol_uninitialized:
+		grad_insist_fail("uninitialized dictionary symbol found!");
+		break;
+			
+	case dict_symbol_attribute:
+		return &sym->v.attr;
+
+	case dict_symbol_alias:
+		return sym->v.alias;
+	}
+}
+
 static grad_dict_attr_t *
 dict_attr_lookup(char *ident)
 {
 	DICT_SYMBOL *sym = grad_sym_lookup(dict_attr_tab, ident);
-	if (sym) {
-		switch (sym->type) {
-		case dict_symbol_uninitialized:
-			grad_insist_fail("grad_sym_lookup returned uninitialized symbol!");
-			break;
-			
-		case dict_symbol_attribute:
-			return &sym->v.attr;
+	if (sym) 
+		return dict_symbol_ptr(sym);
 
-		case dict_symbol_alias:
-			return sym->v.alias;
-		}
-	}
 	return NULL;
 }
 
@@ -100,15 +108,11 @@ free_value(void *ptr, void *closure ARG_UNUSED)
 void
 dict_free()
 {
-        if (dict_attr_tab)
-                grad_symtab_clear(dict_attr_tab);
-        else
-                dict_attr_tab = grad_symtab_create(sizeof(DICT_SYMBOL), NULL);
+	grad_symtab_free(&dict_attr_tab);
         memset(dict_attr_index, 0, sizeof dict_attr_index);
 
 	grad_list_destroy(&dictionary_values, free_value, NULL);
         grad_list_destroy(&dictionary_vendors, free_vendor, NULL);
-        vendorno = 1;
 }
 
 static int
@@ -152,11 +156,11 @@ typedef struct attr_parser_tab ATTR_PARSER_TAB;
 struct attr_parser_tab {
 	ATTR_PARSER_TAB *next;
 	int attr;
-	attr_parser_fp fun;
+	grad_attr_parser_fp fun;
 };
 static ATTR_PARSER_TAB *attr_parser_tab;
 
-static attr_parser_fp
+static grad_attr_parser_fp
 dict_find_parser(int attr)
 {
 	ATTR_PARSER_TAB *ep;
@@ -167,7 +171,7 @@ dict_find_parser(int attr)
 }
 
 void
-dict_register_parser(int attr, attr_parser_fp fun)
+dict_register_parser(int attr, grad_attr_parser_fp fun)
 {
 	ATTR_PARSER_TAB *e = grad_emalloc(sizeof(*e));
 	e->attr = attr;
@@ -176,6 +180,29 @@ dict_register_parser(int attr, attr_parser_fp fun)
 	attr_parser_tab = e;
 }
 
+static void
+parser_tab_init()
+{
+	if (!attr_parser_tab) {
+		/* Register ascend filters */
+		dict_register_parser(242, grad_ascend_parse_filter);
+		dict_register_parser(243, grad_ascend_parse_filter);
+	}
+}	
+
+static void
+parser_tab_free()
+{
+	ATTR_PARSER_TAB *ep = attr_parser_tab;
+
+	while (ep) {
+		ATTR_PARSER_TAB *next = ep->next;
+		grad_free(ep);
+		ep = next;
+	}
+	attr_parser_tab = NULL;
+}
+
 /* **************************************************************************
  * Parser
  */
@@ -186,26 +213,31 @@ static int parse_dict(char *name);
 #define ATTR_VALUE   fv[2]
 #define ATTR_TYPE    fv[3]
 #define ATTR_VENDOR  fv[4]
-#define ATTR_FLAGS   fv[5]
-#define ATTR_ADDITIVITY fv[6]
-#define HAS_VENDOR(c,p)     ((c>=5)&&strcmp(p[4],"-"))
-#define HAS_FLAGS(c,p)      (c==6)
+#define ATTR_FLAGS   ((fc == 6) ? fv[5] : fv[4])
+#define HAS_VENDOR(c,p)     ((c>=5) && strcmp(p[4],"-") && p[4][0] != '[')
+#define HAS_FLAGS(c,p)      ((c==6) || (c==5 && p[4][0] == '['))
 #define VALUE_ATTR   fv[1]
 #define VALUE_NAME   fv[2]
 #define VALUE_NUM    fv[3]
 #define VENDOR_NAME  fv[1]
 #define VENDOR_VALUE fv[2]
 
-static struct keyword type_kw[] = {
-        { "string", TYPE_STRING },
-        { "integer", TYPE_INTEGER },
-        { "ipaddr", TYPE_IPADDR },
-        { "date", TYPE_DATE }
+static grad_keyword_t type_kw[] = {
+        { "string", GRAD_TYPE_STRING },
+        { "integer", GRAD_TYPE_INTEGER },
+        { "ipaddr", GRAD_TYPE_IPADDR },
+        { "date", GRAD_TYPE_DATE }
+};
+
+struct parse_data {
+	int vendor;
+	grad_locus_t begin_locus;
+	int errcnt;
 };
 
 /*ARGSUSED*/
 static int
-_dict_include(int *errcnt, int fc, char **fv, grad_locus_t *loc)
+_dict_include(struct parse_data *pd, int fc, char **fv, grad_locus_t *loc)
 {
         if (nfields(fc, 2, 2, loc)) 
                 return 0;
@@ -219,7 +251,7 @@ parse_flags(char **ptr, int *flags, grad_locus_t *loc)
         int i;
         char *p;
         
-        for (p = *ptr+1, i = 0; i < CF_MAX; i++) {
+        for (p = *ptr+1, i = 0; i < GRAD_CF_MAX; i++) {
                 if (*p == 0) {
                         grad_log_loc(L_ERR, loc, _("missing ]"), *p);
                         return 1;
@@ -227,10 +259,10 @@ parse_flags(char **ptr, int *flags, grad_locus_t *loc)
                 switch (*p++) {
                 case 'C':
                 case 'L':
-                        *flags |= AF_LHS(i);
+                        *flags |= GRAD_AF_LHS(i);
                         break;
                 case '-':
-                        *flags &= ~AF_LHS(i);
+                        *flags &= ~GRAD_AF_LHS(i);
                         break;
                 case ']':
                         p--;
@@ -243,10 +275,10 @@ parse_flags(char **ptr, int *flags, grad_locus_t *loc)
                 }
                 switch (*p++) {
                 case 'R':
-                        *flags |= AF_RHS(i);
+                        *flags |= GRAD_AF_RHS(i);
                         break;
                 case '-':
-                        *flags &= ~AF_RHS(i);
+                        *flags &= ~GRAD_AF_RHS(i);
                         break;
                 default:
                         grad_log_loc(L_ERR, loc,
@@ -256,8 +288,8 @@ parse_flags(char **ptr, int *flags, grad_locus_t *loc)
                 }
         }
   stop:
-        for (; i < CF_MAX; i++) 
-                *flags |= AF_LHS(i)|AF_RHS(i);
+        for (; i < GRAD_CF_MAX; i++) 
+                *flags |= GRAD_AF_LHS(i)|GRAD_AF_RHS(i);
         *ptr = p;
         return 0;
 }
@@ -272,14 +304,14 @@ parse_attr_properties(grad_locus_t *loc, char *str, int *flags, int *prop)
 		switch (*p) {
 		case 'C':
 		case 'L':
-			*flags |= AF_LHS(CF_USERS)
-				   |AF_LHS(CF_HINTS)
-				   |AF_LHS(CF_HUNTGROUPS);
+			*flags |= GRAD_AF_LHS(GRAD_CF_USERS)
+				   |GRAD_AF_LHS(GRAD_CF_HINTS)
+				   |GRAD_AF_LHS(GRAD_CF_HUNTGROUPS);
 			break;
 		case 'R':
-			*flags |= AF_RHS(CF_USERS)
-				   |AF_RHS(CF_HINTS)
-				   |AF_RHS(CF_HUNTGROUPS);
+			*flags |= GRAD_AF_RHS(GRAD_CF_USERS)
+				   |GRAD_AF_RHS(GRAD_CF_HINTS)
+				   |GRAD_AF_RHS(GRAD_CF_HUNTGROUPS);
 			break;
 		case '[':
 			if (parse_flags(&p, flags, loc)) {
@@ -289,19 +321,19 @@ parse_attr_properties(grad_locus_t *loc, char *str, int *flags, int *prop)
 			}
 			break;
 		case '=':
-			SET_ADDITIVITY(*prop, AP_ADD_REPLACE);
+			GRAD_SET_ADDITIVITY(*prop, GRAD_AP_ADD_REPLACE);
 			break;
 		case '+':
-			SET_ADDITIVITY(*prop, AP_ADD_APPEND);
+			GRAD_SET_ADDITIVITY(*prop, GRAD_AP_ADD_APPEND);
 			break;
 		case 'N':
-			SET_ADDITIVITY(*prop, AP_ADD_NONE);
+			GRAD_SET_ADDITIVITY(*prop, GRAD_AP_ADD_NONE);
 			break;
 		case 'P':
-			*prop |= AP_PROPAGATE;
+			*prop |= GRAD_AP_PROPAGATE;
 			break;
 		case 'l':
-			*flags &= ~AP_INTERNAL;
+			*flags &= ~GRAD_AP_INTERNAL;
 			break;
 		case '1':
 		case '2':
@@ -312,16 +344,16 @@ parse_attr_properties(grad_locus_t *loc, char *str, int *flags, int *prop)
 		case '7':
 		case '8':
 		case '9':
-			*prop |= AP_USER_FLAG(*p-'0');
+			*prop |= GRAD_AP_USER_FLAG(*p-'0');
 			break;
 		case 'b':
-			*prop |= AP_BINARY_STRING;
+			*prop |= GRAD_AP_BINARY_STRING;
 			break;
 		case 'E':
-			*prop |= AP_ENCRYPT_RFC2138;
+			*prop |= GRAD_AP_ENCRYPT_RFC2138;
 			break;
 		case 'T':
-			*prop |= AP_ENCRYPT_RFC2868;
+			*prop |= GRAD_AP_ENCRYPT_RFC2868;
 			break;
 		case 'c':
 			/* Retained for compatibility */
@@ -340,22 +372,22 @@ parse_attr_properties(grad_locus_t *loc, char *str, int *flags, int *prop)
 static void
 set_default_attr_properties(int value, int *flags, int *prop)
 {
-        *flags = AF_DEFAULT_FLAGS;
-        *prop  = AP_DEFAULT_ADD;
+        *flags = GRAD_AF_DEFAULT_FLAGS;
+        *prop  = GRAD_AP_DEFAULT_ADD;
 
 	if (GRAD_VENDOR_CODE(value) == 0) {
 		if (value > 255)
-			*flags |= AP_INTERNAL;
+			*flags |= GRAD_AP_INTERNAL;
 		/* FIXME: A temporary hack until all users update
 		   their dictionaries */
 		else if (value == DA_USER_PASSWORD
 			 || value == DA_USER_PASSWORD)
-			*prop |= AP_ENCRYPT_RFC2138;
+			*prop |= GRAD_AP_ENCRYPT_RFC2138;
 	}
 }
 
 static int
-_dict_attribute(int *errcnt, int fc, char **fv, grad_locus_t *loc)
+_dict_attribute(struct parse_data *pd, int fc, char **fv, grad_locus_t *loc)
 {
 	DICT_SYMBOL *sym;
         grad_dict_attr_t *attr;
@@ -363,12 +395,20 @@ _dict_attribute(int *errcnt, int fc, char **fv, grad_locus_t *loc)
         int vendor = 0;
         unsigned value;
         char *p;
-	attr_parser_fp fp = NULL;
+	grad_attr_parser_fp fp = NULL;
         int flags;
         int prop;
         
         if (nfields(fc, 4, 6, loc))
                 return 0;
+
+	debug(100,
+	      ("%s:%lu: ATTR_NAME=%s ATTR_VALUE=%s ATTR_TYPE=%s ATTR_VENDOR=%s ATTR_FLAGS=%s",
+	       loc->file, loc->line,
+	       ATTR_NAME,ATTR_VALUE,ATTR_TYPE,
+	       HAS_VENDOR(fc, fv) ? ATTR_VENDOR : pd->vendor ? "BLOCK" : "N/A",
+	       HAS_FLAGS(fc,fv) ? ATTR_FLAGS : "N/A"));
+	
         /*
          * Validate all entries
          */
@@ -378,12 +418,12 @@ _dict_attribute(int *errcnt, int fc, char **fv, grad_locus_t *loc)
                 grad_log_loc(L_ERR, loc,
 			     _("value not a number (near %s)"),
 			     p);
-                (*errcnt)++;
+                pd->errcnt++;
                 return 0;
         }
 
 	if (strcmp(ATTR_TYPE, "abinary") == 0) {
-		type = TYPE_STRING;
+		type = GRAD_TYPE_STRING;
 		fp = dict_find_parser(value);
 		if (!fp) {
 			grad_log_loc(L_WARN, loc,
@@ -391,30 +431,33 @@ _dict_attribute(int *errcnt, int fc, char **fv, grad_locus_t *loc)
 			return 0;
 		}
 	} else
-		type = grad_xlat_keyword(type_kw, ATTR_TYPE, TYPE_INVALID);
+		type = grad_xlat_keyword(type_kw, ATTR_TYPE, GRAD_TYPE_INVALID);
 	
-        if (type == TYPE_INVALID) {
+        if (type == GRAD_TYPE_INVALID) {
                 grad_log_loc(L_ERR, loc,
 			     "%s",
 			     _("invalid type"));
-                (*errcnt)++;
+                pd->errcnt++;
                 return 0;
         }
 
         if (HAS_VENDOR(fc, fv)) {
                 if ((vendor = grad_vendor_name_to_id(ATTR_VENDOR)) == 0) {
                         grad_log_loc(L_ERR, loc, _("unknown vendor"));
-			(*errcnt)++;
+			pd->errcnt++;
                         return 0;
                 }
-		value |= (vendor << 16);
-        } 
+        } else if (pd->vendor)
+		vendor = pd->vendor;
+
+	value |= (vendor << 16);
+	
 	set_default_attr_properties(value, &flags, &prop);
 
         if (HAS_FLAGS(fc,fv)) {
 		int rc = parse_attr_properties(loc, ATTR_FLAGS, &flags, &prop);
 		if (rc) {
-			++*errcnt;
+			pd->errcnt++;
 			return 0;
 		}
         }
@@ -454,7 +497,7 @@ _dict_attribute(int *errcnt, int fc, char **fv, grad_locus_t *loc)
 /* Syntax:
    ALIAS oldname newname */
 static int
-_dict_alias(int *errcnt, int fc, char **fv, grad_locus_t *loc)
+_dict_alias(struct parse_data *pd, int fc, char **fv, grad_locus_t *loc)
 {
 	DICT_SYMBOL *sym;
 	grad_dict_attr_t *attr;
@@ -488,7 +531,7 @@ _dict_alias(int *errcnt, int fc, char **fv, grad_locus_t *loc)
    
    PROPERTY Attribute Flags */
 static int
-_dict_property(int *errcnt, int fc, char **fv, grad_locus_t *loc)
+_dict_property(struct parse_data *pd, int fc, char **fv, grad_locus_t *loc)
 {
 	grad_dict_attr_t *attr;
 	int i;
@@ -512,7 +555,7 @@ _dict_property(int *errcnt, int fc, char **fv, grad_locus_t *loc)
 			flags = prop = 0;
 			if (parse_attr_properties(loc, fv[i]+1,
 						  &flags, &prop)) {
-				++*errcnt;
+				pd->errcnt++;
 				break;
 			}
 			attr->prop |= flags | prop;
@@ -522,7 +565,7 @@ _dict_property(int *errcnt, int fc, char **fv, grad_locus_t *loc)
 			flags = prop = 0;
 			if (parse_attr_properties(loc, fv[i]+1,
 						  &flags, &prop)) {
-				++*errcnt;
+				pd->errcnt++;
 				break;
 			}
 			attr->prop &= ~(flags | prop);
@@ -532,7 +575,7 @@ _dict_property(int *errcnt, int fc, char **fv, grad_locus_t *loc)
 			if (i > 2) {
 				grad_log_loc(L_ERR, loc,
 					     _("PROPERTY syntax error"));
-				++*errcnt;
+				pd->errcnt++;
 			} else {
 				set_default_attr_properties(attr->value,
 							    &flags, &prop);
@@ -540,7 +583,7 @@ _dict_property(int *errcnt, int fc, char **fv, grad_locus_t *loc)
 							  &flags, &prop) == 0)
 					attr->prop = flags | prop;
 				else
-					++*errcnt;
+					pd->errcnt++;
 				break;
 			}
 		}
@@ -549,7 +592,7 @@ _dict_property(int *errcnt, int fc, char **fv, grad_locus_t *loc)
 }
 
 static int
-_dict_value(int *errcnt, int fc, char **fv, grad_locus_t *loc)
+_dict_value(struct parse_data *pd, int fc, char **fv, grad_locus_t *loc)
 {
         grad_dict_value_t *dval;
         grad_dict_attr_t *attr;
@@ -564,7 +607,7 @@ _dict_value(int *errcnt, int fc, char **fv, grad_locus_t *loc)
                 grad_log_loc(L_ERR, loc,
 			     _("value not a number (near %s)"),
 			     p);
-                (*errcnt)++;
+                pd->errcnt++;
                 return 0;
         }
 
@@ -573,11 +616,11 @@ _dict_value(int *errcnt, int fc, char **fv, grad_locus_t *loc)
 		grad_log_loc(L_ERR, loc,
 			     _("Attribute %s is not defined"),
 			     VALUE_ATTR);
-                (*errcnt)++;
+                pd->errcnt++;
 		return 0;
 	}
 
-	attr->prop |= AP_TRANSLATE;
+	attr->prop |= GRAD_AP_TRANSLATE;
 	
         /* Create a new VALUE entry for the list */
         dval = grad_emalloc(sizeof(grad_dict_value_t));
@@ -595,73 +638,175 @@ _dict_value(int *errcnt, int fc, char **fv, grad_locus_t *loc)
 }
 
 static int
-_dict_vendor(int *errcnt, int fc, char **fv, grad_locus_t *loc)
+_dict_vendor(struct parse_data *pd, int fc, char **fv, grad_locus_t *loc)
 {
         int value;
         char *p;
 
         if (nfields(fc, 3, 3, loc))
-                return 0;
+                return 1;
 
         value = strtol(VENDOR_VALUE, &p, 0);
         if (*p) {
                 grad_log_loc(L_ERR, loc,
 			     _("value not a number (near %s)"),
 			     p);
-                (*errcnt)++;
-                return 0;
+                pd->errcnt++;
+                return 1;
         }
 
         if (addvendor(VENDOR_NAME, value) < 0) {
-                (*errcnt)++;
+                pd->errcnt++;
+		return 1;
         }
 
         return 0;
 
 }
 
+static void
+_dict_begin(struct parse_data *pd, int fc, char **fv, grad_locus_t *loc)
+{
+	if (pd->vendor) {
+		grad_log_loc(L_ERR, loc, _("blocks cannot be nested"));
+		grad_log_loc(L_ERR, &pd->begin_locus,
+			     _("block opened here"));
+		pd->errcnt++;
+		return;
+	}
+
+        if (nfields(fc, 3, 4, loc)) {
+		pd->errcnt++;
+                return;
+	}
+	
+	if (strcmp (fv[1], "VENDOR")) {
+		grad_log_loc(L_ERR, loc,
+			     _("block syntax: expected `VENDOR' but found `%s'"),
+			     fv[1]);
+		pd->errcnt++;
+		return;
+	}
+
+	if (fc == 4 && _dict_vendor(pd, fc-1, fv+1, loc))
+		return;
+			
+	if ((pd->vendor = grad_vendor_name_to_id(fv[2])) == 0) {
+		grad_log_loc(L_ERR, loc, _("unknown vendor"));
+		pd->errcnt++;
+		return;
+	}
+	pd->begin_locus = *loc;
+	debug(1, ("%s:%lu: BEGIN VENDOR %s (%d)",
+		  loc->file, loc->line, fv[2], pd->vendor));
+}
+
+static void
+_dict_end(struct parse_data *pd, int fc, char **fv, grad_locus_t *loc)
+{
+	if (!pd->vendor) {
+		grad_log_loc(L_ERR, loc, _("unexpected END"));
+		pd->errcnt++;
+	}
+	debug(1, ("END VENDOR %d", pd->vendor));
+	pd->vendor = 0;
+}
+
+static void
+_dict_begin_vendor(struct parse_data *pd, int fc, char **fv, grad_locus_t *loc)
+{
+	char *args[4];
+
+        if (nfields(fc, 2, 2, loc)) {
+		pd->errcnt++;
+		return;
+	}
+	
+	args[0] = "BEGIN";
+	args[1] = "VENDOR";
+	args[2] = fv[1];
+	args[3] = NULL;
+	_dict_begin(pd, 3, args, loc);
+}
+
+static void
+_dict_end_vendor(struct parse_data *pd, int fc, char **fv, grad_locus_t *loc)
+{
+	_dict_end(pd, fc, fv, loc);
+}
+
 enum {
         KW_INCLUDE,
         KW_ATTRIBUTE,
 	KW_ALIAS,
         KW_VALUE,
         KW_VENDOR,
-	KW_PROPERTY
+	KW_PROPERTY,
+	KW_BEGIN,
+	KW_BEGIN_VENDOR,
+	KW_END_VENDOR,
+	KW_END
 };
 
-static struct keyword dict_kw[] = {
+static grad_keyword_t dict_kw[] = {
         { "$INCLUDE", KW_INCLUDE },
         { "ATTRIBUTE", KW_ATTRIBUTE },
 	{ "ALIAS", KW_ALIAS },
         { "VALUE", KW_VALUE },
         { "VENDOR", KW_VENDOR },
 	{ "PROPERTY", KW_PROPERTY },
+	{ "BEGIN", KW_BEGIN },
+	{ "END", KW_END },
+	{ "BEGIN-VENDOR", KW_BEGIN_VENDOR },
+	{ "END-VENDOR", KW_END_VENDOR },
         { NULL, 0 }
 };
 
 static int
 parse_dict_entry(void *closure, int fc, char **fv, grad_locus_t *loc)
 {
-	int *errcnt = closure;
+	struct parse_data *pd = closure;
         switch (grad_xlat_keyword(dict_kw, KEYWORD, -1)) {
         case KW_INCLUDE:
-                _dict_include(errcnt, fc, fv, loc);
+                _dict_include(pd, fc, fv, loc);
                 break;
+		
         case KW_ATTRIBUTE:
-                _dict_attribute(errcnt, fc, fv, loc);
+                _dict_attribute(pd, fc, fv, loc);
                 break;
+		
 	case KW_ALIAS:
-		_dict_alias(errcnt, fc, fv, loc);
+		_dict_alias(pd, fc, fv, loc);
 		break;
+		
         case KW_VALUE:
-                _dict_value(errcnt, fc, fv, loc);
+                _dict_value(pd, fc, fv, loc);
                 break;
+		
         case KW_VENDOR:
-                _dict_vendor(errcnt, fc, fv, loc);
+                _dict_vendor(pd, fc, fv, loc);
                 break;
+		
 	case KW_PROPERTY:
-		_dict_property(errcnt, fc, fv, loc);
+		_dict_property(pd, fc, fv, loc);
 		break;
+
+	case KW_BEGIN:
+		_dict_begin(pd, fc, fv, loc);
+		break;
+		
+	case KW_END:
+		_dict_end(pd, fc, fv, loc);
+		break;
+
+	case KW_BEGIN_VENDOR:
+		_dict_begin_vendor(pd, fc, fv, loc);
+		break;
+
+	case KW_END_VENDOR:
+		_dict_end_vendor(pd, fc, fv, loc);
+		break;
+		
         default:
                 grad_log_loc(L_ERR, loc, "%s", _("unknown keyword"));
                 break;
@@ -674,17 +819,26 @@ parse_dict(char *name)
 {
         char *path;
         int   rc;
-        int   errcnt = 0;
+	struct parse_data pd;
 
+	pd.errcnt = 0;
+	pd.vendor = 0;
+	
 	if (name[0] == '/')
 		path = grad_estrdup(name);
 	else
-		path = grad_mkfilename(radius_dir, name);
-        rc = grad_read_raddb_file(path, 1, NULL, parse_dict_entry, &errcnt);
-        if (errcnt)
+		path = grad_mkfilename(grad_config_dir, name);
+	debug(1,("parsing %s", path));
+        rc = grad_read_raddb_file(path, 1, NULL, parse_dict_entry, &pd);
+	if (pd.vendor) {
+		grad_log_loc(L_ERR, &pd.begin_locus, _("BEGIN without END"));
+		pd.vendor = 0;
+		pd.errcnt++;
+	}
+        if (pd.errcnt)
                 grad_log(L_NOTICE,
 		         ngettext("%s: %d error", "%s: %d errors",
-			  	  errcnt), path, errcnt);
+			  	  pd.errcnt), path, pd.errcnt);
         grad_free(path);
         return rc;
 }
@@ -692,15 +846,22 @@ parse_dict(char *name)
 int
 grad_dict_init()
 {
-	if (!attr_parser_tab) {
-		/* Register ascend filters */
-		dict_register_parser(242, grad_ascend_parse_filter);
-		dict_register_parser(243, grad_ascend_parse_filter);
-	}
+	parser_tab_free();
         dict_free();
+
+	parser_tab_init();
+	dict_attr_tab = grad_symtab_create(sizeof(DICT_SYMBOL), NULL);
+        vendorno = 1;
         return parse_dict(RADIUS_DICTIONARY);
 }
 
+void
+grad_dict_free()
+{
+	parser_tab_free();
+        dict_free();
+}
+
 /* **************************************************************************
  * Lookup functions
  */
@@ -716,8 +877,10 @@ struct attr_value {
 };
 
 static int
-attrval_cmp(struct attr_value *av, DICT_SYMBOL *sym)
+attrval_cmp(void *data, grad_symbol_t *s)
 {
+	struct attr_value *av = data;
+	DICT_SYMBOL *sym = (DICT_SYMBOL *) s;
 	
 	if (sym->type == dict_symbol_attribute
 	    && sym->v.attr.value == av->value) {
@@ -874,4 +1037,50 @@ grad_vendor_name_to_id(char *name)
 	vp = grad_list_locate(dictionary_vendors, name, vendor_cmp);
         return vp ? vp->vendorcode : 0;
 }
-        
+
+struct dict_iterator {
+	dict_iterator_fp fp;
+	void *closure;
+};
+
+int
+dict_iter_helper(void *data, grad_symbol_t *symbol)
+{
+	struct dict_iterator *p = data;
+	DICT_SYMBOL *dsym = (DICT_SYMBOL *) symbol;
+	grad_dict_attr_t *attr;
+
+	return p->fp(p->closure, dsym->name, dict_symbol_ptr(dsym));
+}
+	
+void
+grad_dictionary_iterate(dict_iterator_fp fp, void *closure)
+{
+	static struct dict_iterator d;
+	d.fp = fp;
+	d.closure = closure;
+	grad_symtab_iterate(dict_attr_tab, dict_iter_helper, &d);
+}
+
+struct dict_value_iterator {
+	dict_value_iterator_fp fp;
+	void *closure;
+};
+
+static int
+dict_value_iter_helper(void *item, void *data)
+{
+	struct dict_value_iterator *p = data;
+	return p->fp(p->closure, item);
+}
+
+void
+grad_dictionary_value_iterate(dict_value_iterator_fp fp, void *closure)
+{
+	static struct dict_value_iterator d;
+	d.fp = fp;
+	d.closure = closure;
+	grad_list_iterate(dictionary_values, dict_value_iter_helper, &d);
+}
+
+
