@@ -1,5 +1,5 @@
 /* This file is part of GNU Radius.
-   Copyright (C) 2000,2001,2002,2003,2004 Free Software Foundation, Inc.
+   Copyright (C) 2000,2001,2002,2003,2004,2005 Free Software Foundation, Inc.
 
    Written by Sergey Poznyakoff
   
@@ -13,9 +13,10 @@
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
   
-   You should have received a copy of the GNU General Public License
-   along with GNU Radius; if not, write to the Free Software Foundation, 
-   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
+   You should have received a copy of the GNU General Public
+   License along with GNU Radius; if not, write to the Free
+   Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
+   Boston, MA 02110-1301 USA. */
 
 #ifdef HAVE_CONFIG_H
 # include <config.h>
@@ -56,10 +57,10 @@ dump(time_t t)
 	FILE *fp;
 	char *fname;
 	
-	if (!radlog_dir)
+	if (!grad_log_dir)
 		return;
 	asprintf(&fname, "%s/radmem.%lu",
-		 radlog_dir, (unsigned long) getpid());
+		 grad_log_dir, (unsigned long) getpid());
 	fp = fopen(fname, "w");
 	if (!fp) 
 		grad_log(L_ERR|L_PERROR, "cannot open file %s", fname);
@@ -217,6 +218,15 @@ grad_free(void *ptr)
         free(ptr);
 }
 
+void
+grad_destroy(void **pptr)
+{
+	if (*pptr) {
+		grad_free(*pptr);
+		*pptr = NULL;
+	}
+}
+
 void *
 grad_emalloc(size_t size)
 {
